# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# type: ignore

from io import StringIO
from typing import Optional, Sequence, Union

from ..damlast.daml_lf_1 import (
    Block,
    DefDataType,
    DefTemplate,
    Expr,
    ModuleRef,
    PrimType,
    Pure,
    Scenario,
    Type as NewType,
    TypeConName,
    Update,
)
from ..damlast.util import module_name, package_local_name
from ..model.types import (
    SCALAR_TYPE_BOOL,
    SCALAR_TYPE_CHAR,
    SCALAR_TYPE_DATE,
    SCALAR_TYPE_DECIMAL,
    SCALAR_TYPE_INTEGER,
    SCALAR_TYPE_PARTY,
    SCALAR_TYPE_RELTIME,
    SCALAR_TYPE_TEXT,
    SCALAR_TYPE_TIME,
    SCALAR_TYPE_UNIT,
    ContractIdType,
    EnumType,
    ForAllType,
    ListType,
    OptionalType,
    RecordType,
    ScalarType,
    TextMapType,
    Type as OldType,
    TypeApp,
    TypeReference,
    TypeVariable,
    UpdateType,
    VariantType,
    type_dispatch_table,
)
from ._render_base import PrettyPrintBase
from .util import maybe_parentheses

_OldTypePrim = Union[
    ScalarType, ContractIdType, ListType, OptionalType, TextMapType, UpdateType, NewType.Prim
]


class DamlPrettyPrinter(PrettyPrintBase):

    syntax = "daml"

    def lexer(self):
        try:
            from .pygments_daml_lexer import DAMLLexer
        except ImportError:
            return None

        return DAMLLexer()

    def module_indent(self):
        return ""

    def visit_module_ref_start(self, module_ref: "ModuleRef"):
        return f"module {module_name(module_ref)} where\n"

    def visit_def_data_type(self, def_data_type: "DefDataType") -> str:
        return self.visit_def_template(None, def_data_type)

    def visit_def_template(
        self, template: "Optional[DefTemplate]", def_data_type: "DefDataType"
    ) -> str:
        """
        Render a template (or a simple data type).

        :param template:
            The :class:`DefTemplate` to render. If not supplied, only the :class:`DefDataType` is
            rendered.
        :param def_data_type:
            The :class:`DefDataType` to render.
        :return:
            String rendering of the template and/or data type.
        """
        local_name = ".".join(def_data_type.name.segments)

        from ..damlast.expand import ExpandVisitor, SimplifyVisitor

        ex = ExpandVisitor(self.lookup)
        sp = SimplifyVisitor(self.lookup)

        def render(expr: Expr) -> str:
            # return self.visit_expr(expr)
            try:
                e1 = ex.visit_expr(expr)
                e2 = sp.visit_expr(e1)
                return self.visit_expr(e2)
            except:
                return "???"

        with StringIO() as buf:
            if def_data_type.record is not None:
                type_args = "".join(f" {p.var}" for p in def_data_type.params)
                if template is not None:
                    buf.write("template " + local_name)
                else:
                    buf.write(f"data {local_name}{type_args} = {local_name}")
                buf.write("\n  with\n")
                for field in def_data_type.record.fields:
                    buf.write(f"    {field.field}: {self.visit_type(field.type)}\n")
                if template is not None:
                    buf.write("  where\n")
                    buf.write(f"    signatory {render(template.signatories)}\n")
                    buf.write(f"    observer {render(template.observers)}\n")
                    buf.write(f"    ensure {render(template.precond)}\n")
                    buf.write(f"    agreement {render(template.agreement)}\n")
                    # buf.write('    controller')
            elif def_data_type.variant is not None:
                buf.write("data " + local_name + " = ")

            if template is None:
                if def_data_type.serializable:
                    buf.write("\n  deriving (Eq, Show)")

            return buf.getvalue()

    # region visit_expr_* methods

    def visit_expr_prim_con_inline_unit(self):
        return "()"

    def visit_expr_rec_con_decl(self, rec_con: "Expr.RecCon", context: "CodeContext"):
        return self._visit_expr_decl(
            context, self.visit_expr_rec_con_inline(rec_con, line_endings=True)
        )

    def visit_expr_rec_con_inline(self, rec_con: "Expr.RecCon", line_endings: bool = False):
        ctor_expr = f"{self.visit_type_con(rec_con.tycon)} with"
        field_exprs = [f"{fwt.field}={self.visit_expr(fwt.expr)}" for fwt in rec_con.fields]
        if line_endings:
            return ctor_expr + "".join("\n  " + expr for expr in field_exprs)
        else:
            return f'{ctor_expr} {"; ".join(field_exprs)}'

    def visit_expr_rec_proj(self, rec_proj: "Expr.RecProj"):
        rec_text = maybe_parentheses(self.visit_expr(rec_proj.record))
        return f"{rec_text}.{rec_proj.field}"

    def visit_expr_variant_con(self, variant_con: "Expr.VariantCon") -> str:
        arg_text = self.visit_expr(variant_con.variant_arg)
        return f"{variant_con.variant_con} {maybe_parentheses(arg_text)}"

    def visit_expr_struct_proj(self, struct_proj: "Expr.StructProj") -> str:
        tuple_text = maybe_parentheses(self.visit_expr(struct_proj.tuple))
        return f"{tuple_text}.{struct_proj.field}"

    def visit_expr_app_inline(self, app: "Expr.App") -> str:
        components = [app.fun, *app.args]
        return " ".join(maybe_parentheses(self.visit_expr(e)) for e in components)

    def visit_expr_abs_decl(self, abs_: "Expr.Abs", context: "CodeContext"):
        local_name = ".".join(context.decl_name)
        field_def = f"{local_name} :: {self.visit_type(context.decl_type)}"
        all_params = [t.var for t in context.type_abs] if context.type_abs is not None else []
        all_params.extend(p.var for p in abs_.param)

        param_text = "".join(" " + p for p in all_params)
        body_text = self.visit_expr(abs_.body)
        field_decl = f"{local_name}{param_text} = {body_text}"
        return f"{field_def}\n{field_decl}"

    def visit_expr_abs_inline(self, abs_: "Expr.Abs"):
        param_text = " ".join(p.var for p in abs_.param)
        body_expr = self.visit_expr(abs_.body)
        return f"(\\{param_text} -> {body_expr})"

    def visit_expr_case(self, case: "Case"):
        scrut_text = maybe_parentheses(self.visit_expr(case.scrut))
        patterns = ["  " + self.visit_expr_casealt(alt) for alt in case.alts]
        lines = [f"case {scrut_text} of"]
        lines.extend(patterns)
        return "\n".join(lines)

    def visit_expr_let(self, let: "Block") -> str:
        binding_lines = [
            f"{binding.binder.var} = {self.visit_expr(binding.bound)}" for binding in let.bindings
        ]
        with StringIO() as buf:
            buf.write("let ")
            delim = ""
            for line in binding_lines:
                buf.write(delim)
                buf.write(line)
                buf.write("\n")
                delim = "    "
            buf.write("in ")
            expr_text = self.visit_expr(let.body)
            buf.write(expr_text)
            return buf.getvalue()

    def visit_expr_nil_inline(self, nil: "Expr.Nil") -> str:
        return "nil"

    def visit_expr_cons(self, cons: "Expr.Cons") -> str:
        front_texts = [self.visit_expr(front) for front in cons.front]
        if cons.tail.nil:
            return "[" + ", ".join(maybe_parentheses(t, ",") for t in front_texts) + "]"

        tail_text = self.visit_expr(cons.tail)
        if len(front_texts) > 1:
            return (
                "["
                + ", ".join(maybe_parentheses(t, ",") for t in front_texts)
                + "] ++ "
                + maybe_parentheses(tail_text)
            )
        else:
            return f"{front_texts[0]} `Cons` {tail_text}"

    def visit_expr_update(self, update: "Update"):
        return update.Sum_match(
            self.visit_expr_update_pure,
            self.visit_expr_update_block,
            self.visit_expr_update_create,
            self.visit_expr_update_exercise,
            self.visit_expr_update_fetch,
            self.visit_expr_update_get_time,
            self.visit_expr_update_lookup_by_key,
            self.visit_expr_update_fetch_by_key,
            self.visit_expr_update_embed_expr,
        )

    def visit_expr_update_pure(self, pure: "Pure") -> str:
        return self._visit_pure(pure)

    def visit_expr_update_block(self, block: "Block") -> str:
        return self._visit_block(block)

    def visit_expr_update_create(self, create: "Update.Create") -> str:
        return f"create " + self.visit_expr(create.expr)

    def visit_expr_update_exercise(self, exercise: "Update.Exercise") -> str:
        return f"exercise " + self.visit_expr(exercise.cid) + self.visit_expr(exercise.arg)

    def visit_expr_update_fetch(self, fetch: "Update.Fetch") -> str:
        return f"fetch " + self.visit_expr(fetch.cid)

    def visit_expr_update_get_time(self, get_time: "Unit") -> str:
        return "getTime"

    def visit_expr_update_lookup_by_key(self, lookup_by_key: "Update.RetrieveByKey") -> str:
        return (
            "lookupByKey @"
            + self.visit_type(lookup_by_key.template)
            + " "
            + self.visit_expr(lookup_by_key.key)
        )

    def visit_expr_update_fetch_by_key(self, fetch_by_key: "Update.RetrieveByKey") -> str:
        return (
            "fetchByKey @"
            + self.visit_type(fetch_by_key.template)
            + " "
            + self.visit_expr(fetch_by_key.key)
        )

    def visit_expr_update_embed_expr(self, embed_expr: "Update.EmbedExpr") -> str:
        return self.visit_expr(embed_expr.body)

    def _visit_pure(self, pure: "Pure") -> str:
        text = self.visit_expr(pure.expr)
        return f"return $ {text}" if " " in text else f"return {text}"

    def _visit_block(self, block: "Block") -> str:
        from .util import indent

        with StringIO() as buf:
            buf.write("do\n")
            for binding in block.bindings:
                buf.write(
                    f"  {binding.binder.var} <- {indent(self.visit_expr(binding.bound), 2).lstrip()}\n"
                )
            for line in self.visit_expr(block.body).splitlines():
                buf.write(f"  {line}\n")
            return buf.getvalue()

    def visit_expr_scenario(self, scenario: "Scenario") -> str:
        return scenario.Sum_match(
            self.visit_expr_scenario_pure,
            self.visit_expr_scenario_block,
            self.visit_expr_scenario_commit,
            self.visit_expr_scenario_must_fail_at,
            self.visit_expr_scenario_pass,
            self.visit_expr_scenario_get_time,
            self.visit_expr_scenario_get_party,
            self.visit_expr_scenario_embed_expr,
        )

    def visit_expr_scenario_pure(self, pure: "Pure") -> str:
        return self._visit_pure(pure)

    def visit_expr_scenario_block(self, block: "Block") -> str:
        return self._visit_block(block)

    def visit_expr_scenario_commit(self, commit: "Scenario.Commit") -> str:
        text = self.visit_expr(commit.expr)
        return f"commit $ {text}" if " " in text else f"commit {text}"

    def visit_expr_scenario_must_fail_at(self, must_fail_at: "Scenario.Commit") -> str:
        text = self.visit_expr(must_fail_at.expr)
        return f"mustFailAt $ {text}" if " " in text else f"mustFailAt {text}"

    def visit_expr_scenario_pass(self, pass_: "Expr") -> str:
        text = self.visit_expr(pass_)
        return f"pass $ {text}" if " " in text else f"pass {text}"

    def visit_expr_scenario_get_time(self, get_time: "Unit") -> str:
        return "getTime"

    def visit_expr_scenario_get_party(self, get_party: "Expr") -> str:
        text = self.visit_expr(get_party)
        return f"getParty $ {text}" if " " in text else f"getParty {text}"

    def visit_expr_scenario_embed_expr(self, embed_expr: "Scenario.EmbedExpr") -> str:
        return self.visit_expr(embed_expr.body)

    def visit_expr_optional_none(self, optional_none: "Expr.OptionalNone") -> str:
        return f"None"

    def visit_expr_optional_some(self, optional_some: "Expr.OptionalSome") -> str:
        expr_text = self.visit_expr(optional_some.body)
        return f"Some {maybe_parentheses(expr_text)}"

    # endregion

    # region visit_type_* methods

    def visit_type(
        self, type: "Union[str, NewType, OldType, TypeConName]", parenthesize: bool = False
    ) -> str:
        from ..damlast.daml_lf_1 import TypeConName
        from ..model.types import TypeReference

        if isinstance(type, str):
            type_str = type
        elif isinstance(type, NewType):
            type_str = type.Sum_match(
                var=self.visit_type_var,
                con=self.visit_type_con,
                prim=self.visit_type_prim,
                tysyn=self.visit_type_syn,
                forall=self.visit_type_forall,
                tuple=self.visit_type_tuple,
                nat=self.visit_type_nat,
                syn=self.visit_type_syn,
            )
        elif isinstance(type, UpdateType):
            type_str = self.visit_type_prim(type)
        elif isinstance(type, ForAllType):
            type_str = self.visit_type_forall(type)
        elif isinstance(type, OldType):
            type_str = type_dispatch_table(
                on_type_var=self.visit_type_var,
                on_type_ref=self.visit_type_con,
                on_scalar=self.visit_type_prim,
                on_contract_id=self.visit_type_prim,
                on_list=self.visit_type_prim,
                on_optional=self.visit_type_prim,
                on_text_map=self.visit_type_prim,
                on_record=self.visit_type_con,
                on_variant=self.visit_type_con,
                on_enum=self.visit_type_con,
                on_type_app=self._visit_type_app,
                on_unsupported=repr,
            )(type)
        elif isinstance(type, TypeConName):
            type_str = self.visit_type_con(TypeReference(type))
        else:
            raise TypeError(f"A DAML Type is required here (got {type!r} instead")

        return maybe_parentheses(type_str) if parenthesize else type_str

    def visit_type_var(self, var: "Union[str, TypeVariable, NewType.Var]"):
        if isinstance(var, NewType.Var):
            return self._visit_type_app((var.var, *var.args))
        elif isinstance(var, TypeVariable):
            return var.name
        elif isinstance(var, str):
            return var
        else:
            raise TypeError(f"A DAML Type variable is required here (got {var!r} instead")

    def visit_type_con(
        self, con: "Union[TypeReference, RecordType, VariantType, EnumType, NewType.Con]"
    ) -> str:
        if isinstance(con, TypeReference):
            return package_local_name(con.con)
        elif isinstance(con, (RecordType, VariantType, EnumType)):
            if con.name is None:
                raise ValueError("A named Record, Variant, or Enum type is required here")
            return package_local_name(con.name.con)
        elif isinstance(con, NewType.Con):
            return self._visit_type_app((package_local_name(con.tycon), *con.args))
        else:
            raise TypeError(f"A DAML Type constructor is required here (got {con!r} instead")

    def visit_type_prim(self, prim: "_OldTypePrim") -> str:
        prim_type = prim.prim if isinstance(prim, NewType.Prim) else None

        if PrimType.UNIT == prim_type or SCALAR_TYPE_UNIT == prim:
            return "Unit"
        elif PrimType.BOOL == prim_type or SCALAR_TYPE_BOOL == prim:
            return "Bool"
        elif PrimType.INT64 == prim_type or SCALAR_TYPE_INTEGER == prim:
            return "Int"
        elif PrimType.DECIMAL == prim_type or SCALAR_TYPE_DECIMAL == prim:
            return "Decimal"
        elif PrimType.CHAR == prim_type or SCALAR_TYPE_CHAR == prim:
            return "Char"
        elif PrimType.TEXT == prim_type or SCALAR_TYPE_TEXT == prim:
            return "Text"
        elif PrimType.TIMESTAMP == prim_type or SCALAR_TYPE_TIME == prim:
            return "Time"
        elif PrimType.RELTIME == prim_type or SCALAR_TYPE_RELTIME == prim:
            return "RelTime"
        elif PrimType.PARTY == prim_type or SCALAR_TYPE_PARTY == prim:
            return "Party"

        elif PrimType.LIST == prim_type:
            if len(prim.args) == 0:
                return "List"
            else:
                return f"[{self.visit_type(prim.args[0])}]"
        elif isinstance(prim, ListType):
            return f"[{self.visit_type(prim.type_parameter)}]"

        elif PrimType.UPDATE == prim_type:
            return self._visit_type_app(("Update", *prim.args))
        elif isinstance(prim, UpdateType):
            return self._visit_type_app(("Update", prim.type_parameter))

        elif PrimType.SCENARIO == prim_type:
            return self._visit_type_app(("Scenario", *prim.args))

        elif PrimType.DATE == prim_type or SCALAR_TYPE_DATE == prim:
            return "Date"

        elif PrimType.CONTRACT_ID == prim_type:
            return self._visit_type_app(("ContractId", *prim.args))
        elif isinstance(prim, ContractIdType):
            return self._visit_type_app(("ContractId", prim.type_parameter))

        elif PrimType.OPTIONAL == prim_type:
            return self._visit_type_app(("Optional", *prim.args))
        elif isinstance(prim, OptionalType):
            return self._visit_type_app(("Optional", prim.type_parameter))

        elif PrimType.ARROW == prim_type:
            arrow_operator = " -> "
            return arrow_operator.join(
                maybe_parentheses(self.visit_type(a), arrow_operator) for a in prim.args
            )

        elif PrimType.NUMERIC == prim_type:
            return self._visit_type_app(("Numeric", *prim.args))

        elif PrimType.TEXTMAP == prim_type:
            return self._visit_type_app(("TextMap", *prim.args))

        elif PrimType.TYPE_REP == prim.prim:
            return "???"

        elif PrimType.ANY == prim_type:
            return "Any"

        else:
            raise TypeError(f"A DAML Type primitive is required here (got {prim!r} instead")

    def visit_type_forall(self, forall: "Union[ForAllType, NewType.Forall]") -> str:
        if isinstance(forall, NewType.Forall):
            return f'forall {" ".join(self.visit_type_var(v.var) for v in forall.vars)}. {self.visit_type(forall.body)}'
        elif isinstance(forall, ForAllType):
            return (
                f'forall {" ".join(self.visit_type_var(v.var) for v in forall.type_vars)}. '
                f"{self.visit_type(forall.body_type)}"
            )
        else:
            raise TypeError(f"A DAML forall Type is required here (got {forall!r} instead")

    # noinspection PyShadowingBuiltins
    def visit_type_tuple(self, tuple: "NewType.Tuple") -> str:
        return "(" + ",".join(self.visit_type(t.type) for t in tuple.fields) + ")"

    def _visit_type_app(self, types: "Union[TypeApp, Sequence[Union[str, NewType, OldType]]]"):
        if isinstance(types, TypeApp):
            types = (types.body, *types.arguments)

        return " ".join(self.visit_type(c, True) for c in types)

    # endregion


DEFAULT_PRINTER = DamlPrettyPrinter()



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .package_service_pb2 import (
    GetPackageRequest,
    GetPackageResponse,
    GetPackageStatusRequest,
    GetPackageStatusResponse,
    ListPackagesRequest,
    ListPackagesResponse,
)

__all__ = [
    "PackageServiceStub"
]

# noinspection PyPep8Naming
class PackageServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _PackageServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _PackageServiceStub_Async: ...
    def ListPackages(self, __1: ListPackagesRequest) -> __Union[ListPackagesResponse, __Awaitable[ListPackagesResponse]]: ...
    def GetPackage(self, __1: GetPackageRequest) -> __Union[GetPackageRequest, __Awaitable[GetPackageResponse]]: ...
    def GetPackageStatus(self, __1: GetPackageStatusRequest) -> __Union[GetPackageStatusResponse, __Awaitable[GetPackageStatusResponse]]: ...

class _PackageServiceStub(PackageServiceStub):
    def ListPackages(self, __1: ListPackagesRequest) -> ListPackagesResponse: ...
    def GetPackage(self, __1: GetPackageRequest) -> GetPackageResponse: ...
    def GetPackageStatus(self, __1: GetPackageStatusRequest) -> GetPackageStatusResponse: ...

class _PackageServiceStub_Async(PackageServiceStub):
    def ListPackages(self, __1: ListPackagesRequest) -> __Awaitable[ListPackagesResponse]: ...
    def GetPackage(self, __1: GetPackageRequest) -> __Awaitable[GetPackageResponse]: ...
    def GetPackageStatus(self, __1: GetPackageStatusRequest) -> __Awaitable[GetPackageStatusResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    DefaultDict,
    List,
    TypeVar,
    Union,
    overload,
)

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

__all__ = ["QueryStreamBase"]

from ..damlast.daml_lf_1 import TypeConName
from .api_types import ArchiveEvent, Boundary, CreateEvent

AnyHandler = Callable[[Any], None]
AAnyHandler = Callable[[Any], Awaitable[None]]

BoundaryHandler = Callable[[Boundary], None]
ABoundaryHandler = Callable[[Boundary], Awaitable[None]]

CreateEventHandler = Callable[[CreateEvent], None]
ACreateEventHandler = Callable[[CreateEvent], Awaitable[None]]

ArchiveEventHandler = Callable[[ArchiveEvent], None]
AArchiveEventHandler = Callable[[ArchiveEvent], Awaitable[None]]

Self = TypeVar("Self")

class QueryStreamBase:
    @property
    def _callbacks(self) -> DefaultDict[str, List[Union[AnyHandler, AAnyHandler]]]: ...
    @overload
    def on_boundary(self, fn: BoundaryHandler) -> BoundaryHandler: ...
    @overload
    def on_boundary(self, fn: ABoundaryHandler) -> ABoundaryHandler: ...
    @overload
    def on_create(self) -> _CreateEventDecorator: ...
    @overload
    def on_create(self, fn: CreateEventHandler) -> CreateEventHandler: ...
    @overload
    def on_create(self, fn: ACreateEventHandler) -> ACreateEventHandler: ...
    @overload
    def on_create(self, template_id: Union[str, TypeConName]) -> _CreateEventDecorator: ...
    @overload
    def on_create(
        self, template_id: Union[str, TypeConName], fn: CreateEventHandler
    ) -> CreateEventHandler: ...
    @overload
    def on_create(
        self, template_id: Union[str, TypeConName], fn: ACreateEventHandler
    ) -> ACreateEventHandler: ...
    @overload
    def on_archive(self) -> _ArchiveEventDecorator: ...
    @overload
    def on_archive(self, fn: ArchiveEventHandler) -> ArchiveEventHandler: ...
    @overload
    def on_archive(self, fn: AArchiveEventHandler) -> AArchiveEventHandler: ...
    @overload
    def on_archive(self, template_id: Union[str, TypeConName]) -> _ArchiveEventDecorator: ...
    @overload
    def on_archive(
        self, template_id: Union[str, TypeConName], fn: ArchiveEventHandler
    ) -> ArchiveEventHandler: ...
    @overload
    def on_archive(
        self, template_id: Union[str, TypeConName], fn: AArchiveEventHandler
    ) -> AArchiveEventHandler: ...
    async def __aenter__(self: Self) -> Self: ...
    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None: ...
    def creates(self) -> AsyncIterator[CreateEvent]: ...
    def events(self) -> AsyncIterator[Union[CreateEvent, ArchiveEvent]]: ...
    def items(self) -> AsyncIterator[Union[CreateEvent, ArchiveEvent, Boundary]]: ...
    def __aiter__(self) -> AsyncIterator[Union[CreateEvent, ArchiveEvent, Boundary]]: ...
    async def close(self) -> None: ...
    async def run(self) -> None: ...
    async def _emit(self, name: str, obj: Any) -> None: ...
    async def _emit_create(self, event: CreateEvent) -> None: ...
    async def _emit_archive(self, event: ArchiveEvent) -> None: ...
    async def _emit_boundary(self, boundary: Boundary) -> None: ...

class _CreateEventDecorator(Protocol):
    @overload
    def __call__(self, fn: CreateEventHandler) -> CreateEventHandler: ...
    @overload
    def __call__(self, fn: ACreateEventHandler) -> AArchiveEventHandler: ...

class _ArchiveEventDecorator(Protocol):
    @overload
    def __call__(self, fn: ArchiveEventHandler) -> ArchiveEventHandler: ...
    @overload
    def __call__(self, fn: AArchiveEventHandler) -> AArchiveEventHandler: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# This module will likely be deprecated in v8.

from typing import Any, Sequence

from ..damlast.daml_lf_1 import Type, TypeConName
from ..damlast.daml_types import con
from ..damlast.lookup import find_choice
from ..damlast.protocols import SymbolLookup
from ..ledger import (
    Command,
    CreateAndExerciseCommand,
    CreateCommand,
    ExerciseByKeyCommand,
    ExerciseCommand,
)
from ..prim import ContractId
from ..values import ValueMapper

__all__ = [
    "Serializer",
    "AbstractSerializer",
]


class Serializer:
    """
    Serializer interface for objects on the write-side of the API.
    """

    def serialize_value(self, tt: "Type", obj: "Any") -> "Any":
        raise NotImplementedError("serialize_value requires an implementation")

    def serialize_command(self, command: "Any") -> "Any":
        raise NotImplementedError("serialize_command requires an implementation")


class AbstractSerializer(Serializer):
    """
    Implementation of :class:`Serializer` that helps enforce that all possible cases of type
    serialization have been implemented.
    """

    def __init__(self, lookup: "SymbolLookup"):
        self.lookup = lookup

    @property
    def mapper(self) -> "ValueMapper":
        raise NotImplementedError(f"{type(self)}.mapper() must be defined")

    def serialize_value(self, tt: "Type", obj: Any) -> "Any":
        from ..values import Context

        return Context(self.mapper, self.lookup).convert(tt, obj)

    def serialize_commands(self, commands: "Sequence[Command]") -> "Sequence[Any]":
        return [self.serialize_command(cmd) for cmd in commands]

    def serialize_command(self, command: "Command") -> "Any":
        if isinstance(command, CreateCommand):
            name = self.lookup.template_name(command.template_id)
            value = self.serialize_value(con(name), command.payload)
            return self.serialize_create_command(name, value)

        elif isinstance(command, ExerciseCommand):
            template = self.lookup.template(command.contract_id.value_type)
            choice = find_choice(template, command.choice)
            args = self.serialize_value(choice.arg_binder.type, command.argument)
            return self.serialize_exercise_command(command.contract_id, choice.name, args)

        elif isinstance(command, CreateAndExerciseCommand):
            name = self.lookup.template_name(command.template_id)
            template = self.lookup.template(name)
            create_value = self.serialize_value(con(name), command.payload)
            choice = find_choice(template, command.choice)
            choice_args = self.serialize_value(choice.arg_binder.type, command.argument)
            return self.serialize_create_and_exercise_command(
                name, create_value, choice.name, choice_args
            )

        elif isinstance(command, ExerciseByKeyCommand):
            name = self.lookup.template_name(command.template_id)
            template = self.lookup.template(name)
            key_type = template.key
            if key_type is None:
                raise ValueError(f"template {template.tycon} does not have a key")
            key_value = self.serialize_value(key_type.type, command.key)
            choice = find_choice(template, command.choice)
            choice_args = self.serialize_value(choice.arg_binder.type, command.argument)
            return self.serialize_exercise_by_key_command(name, key_value, choice.name, choice_args)

        else:
            raise ValueError(f"unknown Command type: {command!r}")

    def serialize_create_command(self, name: "TypeConName", template_args: "Any") -> "Any":
        raise NotImplementedError("serialize_create_command requires an implementation")

    def serialize_exercise_command(
        self, contract_id: "ContractId", choice_name: str, choice_args: "Any"
    ) -> "Any":
        raise NotImplementedError("serialize_exercise_command requires an implementation")

    def serialize_exercise_by_key_command(
        self, name: "TypeConName", key_arguments: "Any", choice_name: str, choice_arguments: "Any"
    ) -> "Any":
        raise NotImplementedError("serialize_exercise_by_key_command requires an implementation")

    def serialize_create_and_exercise_command(
        self, name: "TypeConName", create_args: "Any", choice_name: str, choice_arguments: "Any"
    ) -> "Any":
        raise NotImplementedError(
            "serialize_create_and_exercise_command requires an implementation"
        )



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    AsyncIterator as __AsyncIterator,
    Iterable as __Stream,
    Iterator as __Iterator,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel, UnaryStreamMultiCallable as __UnaryStreamCall
from grpc.aio import Channel as __AsyncChannel, UnaryStreamCall as __AsyncUnaryStreamCall

from .active_contracts_service_pb2 import GetActiveContractsRequest, GetActiveContractsResponse

__all__ = [
    "ActiveContractsServiceStub",
]

class ActiveContractsServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> __ActiveContractsServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> __ActiveContractsServiceStub_Async: ...
    def GetActiveContracts(self, __1: GetActiveContractsRequest) -> __Union[__ActiveContractsServiceStub_GetActiveContracts_Response, __ActiveContractsServiceStub_GetActiveContracts_AsyncResponse]: ...

class __ActiveContractsServiceStub(ActiveContractsServiceStub):
    def GetActiveContracts(self, __1: GetActiveContractsRequest) -> __ActiveContractsServiceStub_GetActiveContracts_Response: ...

class __ActiveContractsServiceStub_GetActiveContracts_Response(__UnaryStreamCall):
    def __iter__(self) -> __Iterator[GetActiveContractsResponse]: ...

class __ActiveContractsServiceStub_Async(ActiveContractsServiceStub):
    def GetActiveContracts(self, __1: GetActiveContractsRequest) -> __ActiveContractsServiceStub_GetActiveContracts_AsyncResponse: ...

class __ActiveContractsServiceStub_GetActiveContracts_AsyncResponse(__AsyncUnaryStreamCall):
    def __aiter__(self) -> __AsyncIterator[GetActiveContractsResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as _List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    TypeVar as __TypeVar,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.empty_pb2 import Empty
from google.protobuf.message import Message as __Message

from .......damlast.daml_lf_1 import PackageRef
from .......prim import Party

__all__ = [
    "Value",
    "Record",
    "RecordField",
    "Identifier",
    "Variant",
    "Enum",
    "List",
    "Optional",
    "Map",
    "Enum",
    "GenMap",
]

T = __TypeVar("T")

class __List(_List[T]):
    def add(self) -> T: ...

class Value(__Message):
    record: Record
    variant: Variant
    contract_id: str
    list: List
    int64: int
    numeric: str
    text: str
    timestamp: float
    party: Party
    bool: bool
    unit: Empty
    date: int
    optional: Optional
    map: Map
    enum: Enum
    gen_map: GenMap
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, *, record: Record = ...): ...
    @overload
    def __init__(self, *, variant: Variant = ...): ...
    @overload
    def __init__(self, *, contract_id: str = ...): ...
    @overload
    def __init__(self, *, list: List = ...): ...
    @overload
    def __init__(self, *, int64: int = ...): ...
    @overload
    def __init__(self, *, numeric: str = ...): ...
    @overload
    def __init__(self, *, text: str = ...): ...
    @overload
    def __init__(self, *, timestamp: float = ...): ...
    @overload
    def __init__(self, *, party: Party = ...): ...
    @overload
    def __init__(self, *, bool: bool = ...): ...
    @overload
    def __init__(self, *, unit: Empty = ...): ...
    @overload
    def __init__(self, *, date: int = ...): ...
    @overload
    def __init__(self, *, optional: Optional = ...): ...
    @overload
    def __init__(self, *, map: Map = ...): ...
    @overload
    def __init__(self, *, enum: Enum = ...): ...
    @overload
    def __init__(self, *, gen_map: GenMap = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Value) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["Sum", "record", "variant", "contract_id", "list", "int64", "numeric", "text", "timestamp", "party", "bool", "unit", "date", "optional,", "map", "enum", "gen_map"]) -> bool: ...
    def ClearField(self, field_name: __Literal["Sum", "record", "variant", "contract_id", "list", "int64", "numeric", "text", "timestamp", "party", "bool", "unit", "date", "optional,", "map", "enum", "gen_map"]) -> None: ...
    def WhichOneof(self, oneof_group: __Literal["Sum"]) -> __Literal[None, "record", "variant", "contract_id", "list", "int64", "numeric", "text", "timestamp", "party", "bool", "unit", "date", "optional,", "map", "enum", "gen_map"]: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Record(__Message):
    record_id: Identifier
    fields: __List[RecordField]
    def __init__(self, *, record_id: __Optional[Identifier] = ..., fields: __Optional[__Sequence[RecordField]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Record) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["record_id", "fields"]) -> bool: ...
    def ClearField(self, field_name: __Literal["record_id", "fields"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class RecordField(__Message):
    label: str
    value: Value
    def __init__(self, *, label: __Optional[str] = ..., value: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: RecordField) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["label", "value"]) -> bool: ...
    def ClearField(self, field_name: __Literal["label", "value"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Identifier(__Message):
    package_id: PackageRef
    module_name: str
    entity_name: str
    def __init__(self, *, package_id: __Optional[PackageRef] = ..., module_name: __Optional[str] = ..., entity_name: __Optional[str] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Identifier) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["package_id", "module_name", "entity_name"]) -> bool: ...
    def ClearField(self, field_name: __Literal["package_id", "module_name", "entity_name"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Variant(__Message):
    variant_id: Identifier
    constructor: str
    value: Value
    def __init__(self, *, variant_id: __Optional[Identifier] = ..., constructor: __Optional[str] = ..., value: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Variant) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["variant_id", "constructor", "value"]) -> bool: ...
    def ClearField(self, field_name: __Literal["variant_id", "constructor", "value"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Enum(__Message):
    enum_id: Identifier
    constructor: str
    def __init__(self, *, enum_id: __Optional[Identifier] = ..., constructor: __Optional[str] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Enum) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["enum_id", "constructor"]) -> bool: ...
    def ClearField(self, field_name: __Literal["enum_id", "constructor"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class List(__Message):
    elements: __List[Value]
    def __init__(self, *, elements: __Optional[__Sequence[Value]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: List) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["elements"]) -> bool: ...
    def ClearField(self, field_name: __Literal["elements"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Optional(__Message):
    value: Value
    def __init__(self, *, value: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Optional) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["value"]) -> bool: ...
    def ClearField(self, field_name: __Literal["value"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Map(__Message):
    entries: __List[Map.Entry]
    def __init__(self, *, entries: __Optional[__Sequence[Map.Entry]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Map) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["entries"]) -> bool: ...
    def ClearField(self, field_name: __Literal["entries"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...
    class Entry(__Message):
        key: str
        value: Value
        def __init__(self, *, key: __Optional[str] = ..., value: __Optional[Value] = ...): ...
        def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
        def __str__(self) -> str: ...
        def __unicode__(self) -> str: ...
        def MergeFrom(self, other_msg: Map.Entry) -> None: ...
        def Clear(self) -> None: ...
        def SetInParent(self) -> None: ...
        def IsInitialized(self) -> bool: ...
        def ParseFromString(self, serialized: bytes) -> int: ...
        def MergeFromString(self, serialized: bytes) -> int: ...
        def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
        def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
        def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
        def HasField(self, field_name: __Literal["key", "value"]) -> bool: ...
        def ClearField(self, field_name: __Literal["key", "value"]) -> None: ...
        def WhichOneof(self, oneof_group: __Void) -> __Void: ...
        def HasExtension(self, extension_handle: __Any) -> bool: ...
        def ClearExtension(self, extension_handle: __Any) -> None: ...
        def UnknownFields(self) -> __Any: ...
        def DiscardUnknownFields(self) -> None: ...
        def ByteSize(self) -> int: ...
        def _SetListener(self, message_listener: __Any) -> None: ...

class GenMap(__Message):
    entries: __List[Entry]
    def __init__(self, *, entries: __Optional[__Sequence[GenMap.Entry]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GenMap) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["entries"]) -> bool: ...
    def ClearField(self, field_name: __Literal["entries"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...
    class Entry(__Message):
        key: Value
        value: Value
        def __init__(self, *, key: __Optional[Value] = ..., value: __Optional[Value] = ...): ...
        def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
        def __str__(self) -> str: ...
        def __unicode__(self) -> str: ...
        def MergeFrom(self, other_msg: GenMap.Entry) -> None: ...
        def Clear(self) -> None: ...
        def SetInParent(self) -> None: ...
        def IsInitialized(self) -> bool: ...
        def ParseFromString(self, serialized: bytes) -> int: ...
        def MergeFromString(self, serialized: bytes) -> int: ...
        def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
        def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
        def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
        def HasField(self, field_name: __Literal["key", "value"]) -> bool: ...
        def ClearField(self, field_name: __Literal["key", "value"]) -> None: ...
        def WhichOneof(self, oneof_group: __Void) -> __Void: ...
        def HasExtension(self, extension_handle: __Any) -> bool: ...
        def ClearExtension(self, extension_handle: __Any) -> None: ...
        def UnknownFields(self) -> __Any: ...
        def DiscardUnknownFields(self) -> None: ...
        def ByteSize(self) -> int: ...
        def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/transaction_filter.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import value_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/ledger/api/v1/transaction_filter.proto",
    package="com.daml.ledger.api.v1",
    syntax="proto3",
    serialized_options=b"\n\026com.daml.ledger.api.v1B\033TransactionFilterOuterClass\252\002\026Com.Daml.Ledger.Api.V1",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n/com/daml/ledger/api/v1/transaction_filter.proto\x12\x16\x63om.daml.ledger.api.v1\x1a"com/daml/ledger/api/v1/value.proto"\xc4\x01\n\x11TransactionFilter\x12W\n\x10\x66ilters_by_party\x18\x01 \x03(\x0b\x32=.com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry\x1aV\n\x13\x46iltersByPartyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.com.daml.ledger.api.v1.Filters:\x02\x38\x01"F\n\x07\x46ilters\x12;\n\tinclusive\x18\x01 \x01(\x0b\x32(.com.daml.ledger.api.v1.InclusiveFilters"L\n\x10InclusiveFilters\x12\x38\n\x0ctemplate_ids\x18\x01 \x03(\x0b\x32".com.daml.ledger.api.v1.IdentifierBN\n\x16\x63om.daml.ledger.api.v1B\x1bTransactionFilterOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3',
    dependencies=[
        com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2.DESCRIPTOR,
    ],
)


_TRANSACTIONFILTER_FILTERSBYPARTYENTRY = _descriptor.Descriptor(
    name="FiltersByPartyEntry",
    full_name="com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry.value",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=222,
    serialized_end=308,
)

_TRANSACTIONFILTER = _descriptor.Descriptor(
    name="TransactionFilter",
    full_name="com.daml.ledger.api.v1.TransactionFilter",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="filters_by_party",
            full_name="com.daml.ledger.api.v1.TransactionFilter.filters_by_party",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _TRANSACTIONFILTER_FILTERSBYPARTYENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=112,
    serialized_end=308,
)


_FILTERS = _descriptor.Descriptor(
    name="Filters",
    full_name="com.daml.ledger.api.v1.Filters",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="inclusive",
            full_name="com.daml.ledger.api.v1.Filters.inclusive",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=310,
    serialized_end=380,
)


_INCLUSIVEFILTERS = _descriptor.Descriptor(
    name="InclusiveFilters",
    full_name="com.daml.ledger.api.v1.InclusiveFilters",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="template_ids",
            full_name="com.daml.ledger.api.v1.InclusiveFilters.template_ids",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=382,
    serialized_end=458,
)

_TRANSACTIONFILTER_FILTERSBYPARTYENTRY.fields_by_name["value"].message_type = _FILTERS
_TRANSACTIONFILTER_FILTERSBYPARTYENTRY.containing_type = _TRANSACTIONFILTER
_TRANSACTIONFILTER.fields_by_name[
    "filters_by_party"
].message_type = _TRANSACTIONFILTER_FILTERSBYPARTYENTRY
_FILTERS.fields_by_name["inclusive"].message_type = _INCLUSIVEFILTERS
_INCLUSIVEFILTERS.fields_by_name[
    "template_ids"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
DESCRIPTOR.message_types_by_name["TransactionFilter"] = _TRANSACTIONFILTER
DESCRIPTOR.message_types_by_name["Filters"] = _FILTERS
DESCRIPTOR.message_types_by_name["InclusiveFilters"] = _INCLUSIVEFILTERS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TransactionFilter = _reflection.GeneratedProtocolMessageType(
    "TransactionFilter",
    (_message.Message,),
    {
        "FiltersByPartyEntry": _reflection.GeneratedProtocolMessageType(
            "FiltersByPartyEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRANSACTIONFILTER_FILTERSBYPARTYENTRY,
                "__module__": "com.daml.ledger.api.v1.transaction_filter_pb2"
                # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry)
            },
        ),
        "DESCRIPTOR": _TRANSACTIONFILTER,
        "__module__": "com.daml.ledger.api.v1.transaction_filter_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.TransactionFilter)
    },
)
_sym_db.RegisterMessage(TransactionFilter)
_sym_db.RegisterMessage(TransactionFilter.FiltersByPartyEntry)

Filters = _reflection.GeneratedProtocolMessageType(
    "Filters",
    (_message.Message,),
    {
        "DESCRIPTOR": _FILTERS,
        "__module__": "com.daml.ledger.api.v1.transaction_filter_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Filters)
    },
)
_sym_db.RegisterMessage(Filters)

InclusiveFilters = _reflection.GeneratedProtocolMessageType(
    "InclusiveFilters",
    (_message.Message,),
    {
        "DESCRIPTOR": _INCLUSIVEFILTERS,
        "__module__": "com.daml.ledger.api.v1.transaction_filter_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.InclusiveFilters)
    },
)
_sym_db.RegisterMessage(InclusiveFilters)


DESCRIPTOR._options = None
_TRANSACTIONFILTER_FILTERSBYPARTYENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Collection as __Sequence,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.duration_pb2 import Duration
from google.protobuf.message import Message as __Message
from google.protobuf.timestamp_pb2 import Timestamp

from .......prim import Party
from .value_pb2 import Identifier, Record, Value

__all__ = [
    "Commands",
    "Command",
    "CreateCommand",
    "ExerciseCommand",
    "ExerciseByKeyCommand",
    "CreateAndExerciseCommand",
]

class Commands(__Message):
    ledger_id: str
    workflow_id: str
    application_id: str
    command_id: str
    party: Party
    commands: __List[Command]
    deduplication_time: Duration
    min_ledger_time_abs: Timestamp
    min_ledger_time_rel: Timestamp
    act_as: __List[Party]
    read_as: __List[Party]
    def __init__(self, *, ledger_id: __Optional[str] = ..., workflow_id: __Optional[str] = ..., application_id: __Optional[str] = ..., command_id: __Optional[str] = ..., party: __Optional[Party] = ..., commands: __Optional[__Sequence[Command]] = ..., deduplication_time: __Optional[Duration] = ..., min_ledger_time_abs: __Optional[Timestamp] = ..., min_ledger_time_rel: __Optional[Timestamp] = ..., act_as: __Optional[__Sequence[Party]] = ..., read_as: __Optional[__Sequence[Party]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Commands) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def ParseFromString(self, serialized: bytes) -> int: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "workflow_id", "application_id", "command_id", "party", "commands", "deduplication_time", "min_ledger_time_abs", "min_ledger_time_rel", "act_as", "read_as"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "workflow_id", "application_id", "command_id", "party", "commands", "deduplication_time", "min_ledger_time_abs", "min_ledger_time_rel", "act_as", "read_as"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Command(__Message):
    create: CreateCommand
    exercise: ExerciseCommand
    exerciseByKey: ExerciseByKeyCommand
    createAndExercise: CreateAndExerciseCommand
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, *, create: CreateCommand = ...): ...
    @overload
    def __init__(self, *, exercise: ExerciseCommand = ...): ...
    @overload
    def __init__(self, *, exerciseByKey: ExerciseByKeyCommand = ...): ...
    @overload
    def __init__(self, *, createAndExercise: CreateAndExerciseCommand = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Command) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["command", "create", "exercise", "exerciseByKey", "createAndExercise"]) -> bool: ...
    def ClearField(self, field_name: __Literal["command", "create", "exercise", "exerciseByKey", "createAndExercise"]) -> None: ...
    def WhichOneof(self, oneof_group: __Literal["command"]) -> __Literal[None, "create", "exercise", "exerciseByKey", "createAndExercise"]: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class CreateCommand(__Message):
    template_id: Identifier
    create_arguments: Record
    def __init__(self, *, template_id: __Optional[Identifier] = ..., create_arguments: __Optional[Record] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: CreateCommand) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["template_id", "create_arguments"]) -> bool: ...
    def ClearField(self, field_name: __Literal["template_id", "create_arguments"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ExerciseCommand(__Message):
    template_id: Identifier
    contract_id: str
    choice: str
    choice_argument: Value
    def __init__(self, *, template_id: __Optional[Identifier] = ..., contract_id: __Optional[str] = ..., choice: __Optional[str] = ..., choice_argument: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ExerciseCommand) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["template_id", "contract_id", "choice", "choice_argument"]) -> bool: ...
    def ClearField(self, field_name: __Literal["template_id", "contract_id", "choice", "choice_argument"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ExerciseByKeyCommand(__Message):
    template_id: Identifier
    contract_key: Value
    choice: str
    choice_argument: Value
    def __init__(self, *, template_id: __Optional[Identifier] = ..., contract_key: __Optional[Value] = ..., choice: __Optional[str] = ..., choice_argument: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ExerciseByKeyCommand) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["template_id", "contract_key", "choice", "choice_argument"]) -> bool: ...
    def ClearField(self, field_name: __Literal["template_id", "contract_key", "choice", "choice_argument"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class CreateAndExerciseCommand(__Message):
    template_id: Identifier
    create_arguments: Record
    choice: str
    choice_argument: Value
    def __init__(self, *, template_id: __Optional[Identifier] = ..., create_arguments: __Optional[Record] = ..., choice: __Optional[str] = ..., choice_argument: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: CreateAndExerciseCommand) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["template_id", "create_arguments", "choice", "choice_argument"]) -> bool: ...
    def ClearField(self, field_name: __Literal["template_id", "create_arguments", "choice", "choice_argument"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

__all__ = ["ContractMatch", "is_match", "Query"]

from .query import ContractMatch, Query, is_match



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import TYPE_CHECKING, Any

from ..damlast.daml_lf_1 import DefDataType, Type

if TYPE_CHECKING:
    from .context import Context

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

__all__ = ["ValueMapper"]


class ValueMapper(Protocol):
    """
    Protocol for an object that can map one representation of DAML-LF values to another
    representation. This is the protocol that wire encoders/decoders adhere to, as well as certain
    types of validators.

    To use an instance of a :class:`ValueMapper` to transform an object, create a :class:`Context`
    with an instance of a :class:`ValueMapper`; then call :meth:`Context.convert`.
    """

    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        raise NotImplementedError("data_record requires an implementation")

    def data_variant(
        self, context: "Context", dt: "DefDataType", variant: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        raise NotImplementedError("data_variant requires an implementation")

    def data_enum(
        self,
        context: "Context",
        dt: "DefDataType",
        enum: "DefDataType.EnumConstructors",
        obj: "Any",
    ) -> "Any":
        raise NotImplementedError("data_variant requires an implementation")

    def prim_unit(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_unit requires an implementation")

    def prim_bool(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_bool requires an implementation")

    def prim_int64(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_int64 requires an implementation")

    def prim_text(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_text requires an implementation")

    def prim_timestamp(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_timestamp requires an implementation")

    def prim_party(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_party requires an implementation")

    def prim_list(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        raise NotImplementedError("prim_list requires an implementation")

    def prim_date(self, context: "Context", obj: "Any") -> "Any":
        raise NotImplementedError("prim_date requires an implementation")

    def prim_contract_id(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        raise NotImplementedError("prim_contract_id requires an implementation")

    def prim_optional(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        raise NotImplementedError("prim_optional requires an implementation")

    def prim_text_map(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        raise NotImplementedError("prim_text_map requires an implementation")

    def prim_numeric(self, context: "Context", nat: int, obj: "Any") -> "Any":
        raise NotImplementedError("prim_numeric requires an implementation")

    def prim_gen_map(
        self, context: "Context", key_type: "Type", value_type: "Type", obj: "Any"
    ) -> "Any":
        raise NotImplementedError("prim_gen_map requires an implementation")



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as _List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    TypeVar as __TypeVar,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.message import Message as __Message
from google.protobuf.timestamp_pb2 import Timestamp

__all__ = [
    "ListKnownPackagesRequest",
    "ListKnownPackagesResponse",
    "PackageDetails",
    "UploadDarFileRequest",
    "UploadDarFileResponse",
]

class ListKnownPackagesRequest(__Message):
    def __init__(self) -> None: ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ListKnownPackagesRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Void) -> __Void: ...
    def ClearField(self, field_name: __Void) -> __Void: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ListKnownPackagesResponse(__Message):
    package_details: PackageDetails
    def __init__(self, *, package_details: __Optional[PackageDetails] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ListKnownPackagesResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["package_details"]) -> bool: ...
    def ClearField(self, field_name: __Literal["package_details"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class PackageDetails(__Message):
    package_id: str
    package_size: int
    known_since: Timestamp
    source_description: str
    def __init__(self, *, package_id: __Optional[str] = ..., package_size: __Optional[int] = ..., known_since: __Optional[Timestamp] = ..., source_description: __Optional[str] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: PackageDetails) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["package_id", "package_size", "known_since", "source_description"]) -> bool: ...
    def ClearField(self, field_name: __Literal["package_id", "package_size", "known_since", "source_description"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class UploadDarFileRequest(__Message):
    dar_file: bytes
    submission_id: str
    def __init__(self, *, dar_file: __Optional[bytes] = ..., submission_id: __Optional[str] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: UploadDarFileRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["dar_file", "submission_id"]) -> bool: ...
    def ClearField(self, field_name: __Literal["dar_file", "submission_id"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class UploadDarFileResponse(__Message):
    def __init__(self) -> None: ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: UploadDarFileResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Void) -> __Void: ...
    def ClearField(self, field_name: __Void) -> __Void: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Simple methods for instantiating an application written against dazl that incorporate common usage
patterns.
"""
from argparse import ArgumentParser
from typing import Callable
import warnings

from ..client.api import Network
from ..client.config import NetworkConfig, configure_parser


def run(init: "Callable[[Network], None]"):
    warnings.warn("dazl.run is deprecated", DeprecationWarning, stacklevel=2)

    if init is None:
        raise ValueError("The init callback cannot be None")

    arg_parser = ArgumentParser()

    configure_parser(arg_parser, config_file_support=True)
    args = arg_parser.parse_args()

    config = NetworkConfig.get_config(args)
    network = Network()
    network.set_config(config)

    init(network)

    network.run_forever()



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import (
    CancelledError,
    Future,
    InvalidStateError,
    Task,
    ensure_future,
    get_event_loop,
    sleep,
)
from collections.abc import Mapping as MappingBase
import sys
from types import MappingProxyType
from typing import (
    AsyncIterator,
    Awaitable,
    Collection,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)
import warnings

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.lookup import matching_normalizations, normalize
from ..ledger import ArchiveEvent, Boundary, CreateEvent
from ..ledger.errors import ProtocolWarning
from ..ledger.grpc.conn_aio import Connection
from ..prim import ContractData, ContractId
from ..query import ContractMatch, is_match

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

__all__ = ["ACS", "Snapshot"]


T = TypeVar("T")

# Literal types are a bit of a pain to declare, but we _do_ get type safety at the value level
# with all this boilerplate.
NOT_STARTED: Literal["NOT_STARTED"] = "NOT_STARTED"
CONNECTING: Literal["CONNECTING"] = "CONNECTING"
RUNNING: Literal["RUNNING"] = "RUNNING"
RECONNECTING: Literal["RECONNECTING"] = "RECONNECTING"
STOPPED: Literal["STOPPED"] = "STOPPED"
State = Literal["NOT_STARTED", "CONNECTING", "RUNNING", "RECONNECTING", "STOPPED"]


class ACS:
    """
    An accumulation of active contracts up to a certain point in time over a provided stream.

    This class is meant to represent a _local_ in-memory view of an active contract set. It is _not_
    a direct wrapper over the API; it is merely offered as a simple utility class to avoid
    repetitive code that looks like this:

    .. code-block: python

        class MyService:
            def __init__(self, conn):
                self._conn = conn
                self._widgets = dict()

            async def run(self):
                async with self._conn.stream("WidgetMod:Widget") as stream:
                    async for ev in stream.events():
                        if isinstance(ev, CreateEvent):
                            self._widgets[ev.contract_id] = ev.payload
                        elif isinstance(ev, ArchiveEvent):
                            self._widgets.pop(ev.contract_id)

            def widget_json(self, widget_id: str):
                for cdata in self._widgets.values():
                    if cdata['id'] == widget_id:
                        return cdata
                return None

    With the ACS helper class:

    .. code-block: python

        class MyService:
            def __init__(self, conn):
                self._conn = conn
                self._acs = ACS(conn, {"WidgetMod:Widget": {}})

            async def run(self):
                await self._acs.run()

            def widget_json(self, widget_id: str):
                snapshot = await self._acs.read()
                return snapshot.earliest_contract("WidgetMod:Widget", {"id": widget_id})

    """

    def __init__(
        self,
        conn: Connection,
        queries: Mapping[Union[str, TypeConName], ContractMatch],
    ):
        self._conn = conn
        self._queries = queries
        self._snapshot = None  # type: Optional[Snapshot]
        self._snapshot_fut = None  # type: Optional[Future]
        self._task = None  # type: Optional[Future]
        self._state = NOT_STARTED  # type: State
        self._log = conn.config.logger

    def read(self) -> "Awaitable[Snapshot]":
        """
        Returns a snapshot of data.

        Despite being asynchronous, this call will generally return immediately unless:
        * :meth:`start` has not yet been called
        * the stream is still reading the initial Active Contract Set
        """
        if self._state == STOPPED:
            raise InvalidStateError("this ACS has already been cancelled")

        if self._snapshot_fut is None:
            self._snapshot_fut = get_event_loop().create_future()
        return self._snapshot_fut

    def read_immediately(self) -> "Optional[Snapshot]":
        """
        Returns the most recent snapshot of data, or ``None`` if this ACS was never started.

        This can be used as an alternative to :meth:`read` when the caller more interested in a
        non-``async`` call, and can tolerate reading data that is potentially slightly behind (or
        not yet present).

        :return: The last good ``Snapshot``, or ``None`` if the stream has not yet been started.
        :raises InvalidStateError: if the stream has already been stopped
        """
        if self._state == STOPPED:
            raise InvalidStateError("this ACS has already been cancelled")

        # if our Snapshot future is in a failed state, this method also throws in order to make
        # any read() exceptions known
        if self._snapshot_fut is not None:
            ex = self._snapshot_fut.exception()
            if ex is not None:
                raise ex

        return self._snapshot

    @property
    def state(self) -> State:
        """
        Return the current connection state:

        * ``NOT_STARTED``: :class:`ACS` has been created, but :meth:`run` has not yet been called.
        * ``CONNECTING``: :meth:`run` has been called, but the Active Contract Set has not yet been
          fully downloaded.
        * ``RUNNING``: the active contract set has been downloaded, and transaction event updates
          will be streamed as they are available.
        * ``RECONNECTING``: the stream was disconnected by the server;
        """
        return self._state

    async def start(self) -> None:
        """
        Have the ACS start subscribing for events from the server. This coroutine "blocks" until a
        snapshot is available.
        """
        if self._state != NOT_STARTED:
            raise InvalidStateError("this ACS is already running")

        # ensure that we "start" our read before we even connect to the stream, so that we can be
        # sure the future is triggered on the very first read
        fut = self.read()
        self._state = CONNECTING
        self._task = ensure_future(self._main())
        await fut

    async def stop(self) -> None:
        """
        Stop the underlying ACS stream. After this point, the stream is considered inoperable.
        """
        if self._task is not None:
            self._task.cancel()
            self._task = None
            self._state = STOPPED
        else:
            raise InvalidStateError("this ACS has already been stopped")

    async def _main(self) -> None:
        """
        The main coroutine. Calls :func:`snapshots` and accumulates relevant state.

        This coroutine NEVER throws because it is never actually awaited; exceptions that occur
        in the body of this method are propagated through the future returned by ``read()``.
        """
        try:
            log = self._conn.config.logger

            log.debug("ACS._main started.")
            async for state, snapshot in snapshots(self._conn, self._queries):
                log.debug("ACS transition: %s, %s", state, snapshot)
                self._state = state

                # any state change causes us to "reset" our read() Future
                if self._snapshot_fut is None or self._snapshot_fut.done():
                    self._snapshot_fut = get_event_loop().create_future()

                if snapshot is not None:
                    self._snapshot = snapshot
                    self._snapshot_fut.set_result(snapshot)
        except Exception as ex:
            # No one is actually waiting for our result, so if we don't move our exception to
            # an awaitable that people actually care about, it will never be seen!
            if self._snapshot_fut is None or self._snapshot_fut.done():
                self._snapshot_fut = get_event_loop().create_future()
            self._snapshot_fut.set_exception(ex)

    async def __aenter__(self) -> "ACS":
        """
        Allow :class:`ACS` to be used within ``async with`` blocks.
        """
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Allow :class:`ACS` to be used within ``async with`` blocks.
        """
        await self.stop()

    def __repr__(self):
        return f"ACS(state={self._state})"


async def snapshots(conn: Connection, queries) -> "AsyncIterator[Tuple[State, Optional[Snapshot]]]":
    """
    Coroutine that returns regular "state" and "snapshot" updates aggregated over events off an
    event stream.
    """
    log = conn.config.logger

    yield CONNECTING, None

    snapshot = None  # type: Optional[Snapshot]
    offset = None  # type: Optional[str]
    recent_events = []  # type: List[Union[CreateEvent, ArchiveEvent]]
    backoff_time = 1.0
    at_least_one_snapshot = False

    while True:
        try:
            async with conn.stream_many(queries, offset=offset) as stream:
                async for obj in stream:
                    # reset the backoff time for subsequent stream disconnect/reconnects
                    backoff_time = 1.0

                    if isinstance(obj, Boundary):
                        offset = obj.offset
                        snapshot = create_snapshot(snapshot, recent_events, offset)

                        recent_events = []
                        at_least_one_snapshot = True
                        yield RUNNING, snapshot

                    elif isinstance(obj, (CreateEvent, ArchiveEvent)):
                        # Upon receiving at least one event, we're going to generate a new snapshot;
                        # inform the caller that a new snapshot is coming. These notifications are
                        # only posted once we have had at least one successful snapshot.
                        if not recent_events and at_least_one_snapshot:
                            yield RUNNING, None

                        # accumulate apps
                        recent_events.append(obj)
                    else:
                        warnings.warn(f"Unrecognized stream element: {obj!r}", ProtocolWarning)

        except CancelledError:
            # once we have been told to cancel, the stream is in the STOPPED state, and no further
            # interactions with it will be successful
            log.info("The ACS stream was asked to stop.")
            yield STOPPED, None
            raise

        except Exception:
            yield RECONNECTING, None
            log.exception(
                "The stream disconnected. Waiting %s seconds, then trying again (attempting "
                "to resume at offset %s)",
                backoff_time,
                offset,
            )
            await sleep(backoff_time)
            backoff_time = max(30.0, backoff_time * 2.0)


class Snapshot:
    """
    An immutable snapshot of contract data.
    """

    def __init__(self, creates: Mapping[ContractId, CreateEvent], offset: Optional[str]):
        self._offset = offset
        self._creates = MappingProxyType(dict(creates))
        self._contracts = ContractDataView(self._creates)
        self._template_ids = {
            cid.value_type: matching_normalizations(cid.value_type) for cid in self._creates
        }

    @property
    def offset(self) -> Optional[str]:
        """
        The offset point in a stream that contains all of the creates without offsetting archives
        in this snapshot.

        This value is ``None`` when connecting to a ledger that is completely empty.
        """
        return self._offset

    def earliest_contract(
        self, template_id: Union[str, TypeConName], match: Optional[ContractMatch] = None
    ) -> Optional[ContractData]:
        """
        Return the earliest contract in the Active Contract Set (in other words, _still_ active)
        that was created in the transaction stream that matches the specified filter, or ``None``
        if there are no matches.
        """
        ev = self.earliest_create(template_id, match)
        return ev.payload if ev is not None else None

    def matching_contracts(
        self, template_id: Union[str, TypeConName], match: Optional[ContractMatch] = None
    ) -> Mapping[ContractId, ContractData]:
        """"""
        return ContractDataView(self.matching_creates(template_id, match))

    def latest_contract(
        self, template_id: Union[str, TypeConName], match: Optional[ContractMatch] = None
    ) -> Optional[ContractData]:
        """
        Return the contract that was created last in the transaction stream that matches the
        specified filter, or ``None`` if there are no matches.
        """
        ev = self.latest_create(template_id, match)
        return ev.payload if ev is not None else None

    def earliest_create(
        self, template_id: Union[str, TypeConName], match: Optional[ContractMatch] = None
    ) -> Optional[CreateEvent]:
        """
        Return the earliest :class:`CreateEvent` in the Active Contract Set (in other words, the
        corresponding contract is _still_ active) that was created in the transaction stream that
        matches the specified filter, or ``None``
        if there are no matches.
        """
        wanted_template_ids = self._matching_template_ids(template_id)

        # in Python 3.9, dict values can be `reversed`; sadly we're not there yet
        for cev in reversed(list(self._creates.values())):
            if cev.contract_id.value_type in wanted_template_ids and is_match(match, cev.payload):
                return cev
        return None

    def matching_creates(
        self, template_id: Union[str, TypeConName], match: Optional[ContractMatch] = None
    ) -> Mapping[ContractId, CreateEvent]:
        """
        Return the :class:`CreateEvent`s (indexed by :class:`ContractId`) whose contracts match the
        specified criteria.
        """
        wanted_template_ids = self._matching_template_ids(template_id)
        matches = {}
        for cid, cev in self._creates.items():
            if cev.contract_id.value_type in wanted_template_ids and is_match(match, cev.payload):
                matches[cev.contract_id] = cev
        return MappingProxyType(matches)

    def latest_create(
        self, template_id: Union[str, TypeConName], match: Optional[ContractMatch] = None
    ) -> Optional[CreateEvent]:
        """
        Return the contract that was created last in the transaction stream that matches the
        specified filter, or ``None`` if there are no matches.
        """
        wanted_template_ids = self._matching_template_ids(template_id)
        for cev in self._creates.values():
            if cev.contract_id.value_type in wanted_template_ids and is_match(match, cev.payload):
                return cev
        return None

    def _matching_template_ids(
        self, template_id: Union[str, TypeConName]
    ) -> Collection[TypeConName]:
        t = normalize(template_id)
        return {tid for tid, matches in self._template_ids.items() if t in matches}

    @property
    def contracts(self) -> Mapping[ContractId, ContractData]:
        """
        A read-only map of contract IDs to contract data.
        """
        return self._contracts

    @property
    def creates(self) -> Mapping[ContractId, CreateEvent]:
        """
        A map of contract IDs to :class:`CreateEvent`s that represent the current set of contracts
        in this ACS.

        :class:`CreateEvent`'s expose additional information, including signatories and observers.
        """
        return self._creates

    def __len__(self):
        return len(self._creates)

    def __repr__(self):
        return f"Snapshot(len={len(self)})"


def create_snapshot(
    base_snapshot: Optional[Snapshot],
    events: Sequence[Union[CreateEvent, ArchiveEvent]],
    offset: Optional[str],
) -> Snapshot:
    """
    Produces a new snapshot with the data from an original snapshot and the supplied events.
    """
    snapshot_data: Dict[ContractId, CreateEvent] = (
        dict(base_snapshot.creates) if base_snapshot is not None else {}
    )
    for event in events:
        if isinstance(event, CreateEvent):
            snapshot_data[event.contract_id] = event
        elif isinstance(event, ArchiveEvent):
            snapshot_data.pop(event.contract_id, None)
    return Snapshot(snapshot_data, offset)


class ContractDataView(MappingBase):
    """
    A map of :class:`ContractId` -> ``ContractData``. This class is just a view over an instance of
    ``Mapping[:class:`ContractId`, :class:`CreateEvent`].
    """

    __slots__ = ("__m",)

    def __init__(self, m: Mapping[ContractId, CreateEvent]):
        self.__m = m

    def __getitem__(self, key: ContractId) -> ContractData:
        return self.__m[key].payload

    def __len__(self) -> int:
        return len(self.__m)

    def __iter__(self) -> Iterator[ContractId]:
        return iter(self.__m)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from os import PathLike, fspath
from pathlib import Path
from typing import Union
import warnings

__all__ = ["pathify"]


def pathify(path: "Union[str, Path, PathLike]") -> Path:
    """
    Convert an object that could be a :class:`Path` into a :class:`Path`.
    """
    warnings.warn("pathify is deprecated; use os.fspath instead", DeprecationWarning, stacklevel=2)
    return Path(fspath(path))



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
:mod:`dazl.ledger` — gRPC Ledger API / HTTP JSON API client
===========================================================
.. autofunction:: connect

This module contains the types needed to submit commands to and read events from a
Daml `gRPC Ledger API <https://docs.daml.com/app-dev/ledger-api.html>`_ or
`HTTP JSON API <https://docs.daml.com/json-api/index.html>`_.

Example:

+-----------------------------------+--------------------------------------------------------------+
| Commands                                                                                         |
+-----------------------------------+--------------------------------------------------------------+
| :class:`Command`                  | abstract base class of all commands                          |
+-----------------------------------+--------------------------------------------------------------+
| :class:`CreateCommand`            | create a contract                                            |
+-----------------------------------+--------------------------------------------------------------+
| :class:`CreateAndExerciseCommand` | create a contract and immediately exercise a choice on the   |
|                                   | newly created contract in a single transaction               |
+-----------------------------------+--------------------------------------------------------------+
| :class:`ExerciseCommand`          | exercise a choice on a contract identified by its contract   |
|                                   | ID                                                           |
+-----------------------------------+--------------------------------------------------------------+
| :class:`ExerciseByKeyCommand`     | exercise a choice on a contract identified by its contract   |
|                                   | key                                                          |
+-----------------------------------+--------------------------------------------------------------+
| Events and Responses                                                                             |
+-----------------------------------+--------------------------------------------------------------+
| :class:`CreateEvent`              | event raised when a contract is created, either as part of   |
|                                   | the active contract set or on the transaction stream         |
+-----------------------------------+--------------------------------------------------------------+
| :class:`ArchiveEvent`             | event raised when a contract is archived on the transaction  |
|                                   | stream                                                       |
+-----------------------------------+--------------------------------------------------------------+
| :class:`ExerciseResponse`         | the response from an exercise, including the choice's return |
|                                   | value (if any)                                               |
+-----------------------------------+--------------------------------------------------------------+
| :class:`Boundary`                 | indicates a point where an event stream can be subsequently  |
|                                   | resumed                                                      |
+-----------------------------------+--------------------------------------------------------------+
| Other                                                                                            |
+-----------------------------------+--------------------------------------------------------------+
| :class:`PartyInfo`                | metadata about a party                                       |
+-----------------------------------+--------------------------------------------------------------+

Write-side types
----------------

Daml ledger state can be updated by submitting commands to the ledger. You should prefer using
:meth:`Connection.create`, :meth:`Connection.create_and_exercise`, :meth:`Connection.exercise`, and
:meth:`Connection.exercise_by_key` over constructing instances of these commands, as those methods
can give you more information about what happened.

.. autoclass:: Command

.. autoclass:: CreateCommand
   :members:

.. autoclass:: CreateAndExerciseCommand
   :members:

.. autoclass:: ExerciseCommand
   :members:

.. autoclass:: ExerciseByKeyCommand
   :members:

Read-side types
---------------

An ``Event`` in a transaction is either a :class:`CreateEvent` or an `ArchiveEvent``.

You can resume a stream of events from a previous point by using ``stream.items()`` and looking for
``Boundary`` objects.

More detailed information about exercises, including the return value and events that occurred as
a result of an exercise, is available from ``ExerciseResponse`` objects as returned from
:meth:`Connection.create_and_exercise`, :meth:`Connection.exercise`, and
:meth:`Connection.exercise_by_key`.

.. autoclass:: CreateEvent
   :members:

.. autoclass:: ArchiveEvent
   :members:

.. autoclass:: Boundary
   :members:

.. autoclass:: ExerciseResponse
   :members:

Other
-----

.. autoclass:: PartyInfo
    :members:

"""
from logging import Logger
from os import PathLike
from typing import Collection, Optional, Union

from ..prim import Party, TimeDeltaLike
from .api_types import (
    ArchiveEvent,
    Boundary,
    Command,
    CreateAndExerciseCommand,
    CreateCommand,
    CreateEvent,
    ExerciseByKeyCommand,
    ExerciseCommand,
    ExerciseResponse,
    PartyInfo,
)

__all__ = [
    "ArchiveEvent",
    "Boundary",
    "Command",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "CreateEvent",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "ExerciseResponse",
    "PartyInfo",
]


def connect(
    url: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    scheme: Optional[str] = None,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = False,
    ledger_id: Optional[str] = None,
    application_name: Optional[str] = None,
    oauth_token: Optional[str] = None,
    oauth_token_file: Optional[str] = None,
    ca: Optional[bytes] = None,
    ca_file: Optional[PathLike] = None,
    cert: Optional[bytes] = None,
    cert_file: Optional[PathLike] = None,
    cert_key: Optional[bytes] = None,
    cert_key_file: Optional[PathLike] = None,
    connect_timeout: Optional[TimeDeltaLike] = None,
    use_http_proxy: bool = True,
    logger: Optional[Logger] = None,
    logger_name: Optional[str] = None,
    log_level: Optional[str] = None,
):
    """
    Create a connection from the supplied parameters.

    The remote can be configured either by supplying a URL or by supplying a host (and optional
    port and scheme). If none of ``url``, ``host``, ``port``, or ``scheme`` are supplied, then
    environment variables are consulted; if no environment variables are specified either, then
    default values are used.

    The remote can be configured either by supplying a URL or by supplying a host (and optional
    port and scheme). If none of ``url``, ``host``, ``port``, or ``scheme`` are supplied, then
    environment variables are consulted; if no environment variables are specified either, then
    default values are used.

    When the URL scheme is ``http`` or ``https``, dazl will first attempt to connect assuming
    the HTTP JSON API; if this fails, gRPC Ledger API is attempted.

    If ``oauth_token`` is supplied and non-empty, then *token-based access* is used to connect
    to the ledger. (Note that you cannot specify both ``oauth_token`` and ``oauth_token_file``.)

    If other fields are specified, then *property-based access* is used. At least one of
    ``read_as``, ``act_as``, or ``admin`` must be supplied. For the HTTP JSON API,
    ``ledger_id`` MUST be supplied.

    If _no_ fields are specified, this is an error unless environment variables supply an
    alternate source of configuration.

    +----------------------+----------------------+------------------------------------------------+
    | function argument    | environment variable | default value                                  |
    +======================+======================+================================================+
    | **Connection config**                                                                        |
    |                                                                                              |
    | Specifying *any* of ``url``, ``host``, ``port``, or ``scheme`` causes *all* of these         |
    | environment variables to be ignored                                                          |
    +----------------------+----------------------+------------------------------------------------+
    | ``url``              | DAML_LEDGER_URL      | localhost:6865                                 |
    +----------------------+----------------------+------------------------------------------------+
    | ``host``             | DAML_LEDGER_HOST     | localhost                                      |
    +----------------------+----------------------+------------------------------------------------+
    | ``port``             | DAML_LEDGER_PORT     | | 6865, unless ``scheme`` is specified:        |
    |                      |                      | | • 80 for ``http``                            |
    |                      |                      | | • 443 for ``https``                          |
    |                      |                      | | • 6865 for ``grpc``                          |
    +----------------------+----------------------+------------------------------------------------+
    | ``scheme``           | DAML_LEDGER_SCHEME   | | ``https`` for port 443 or 8443               |
    |                      |                      | | ``http`` for port 80, 7575 or 8080           |
    |                      |                      | | ``grpc`` for port 6865                       |
    |                      |                      | | ``https`` for all other ports                |
    +----------------------+----------------------+------------------------------------------------+
    | **Access config**                                                                            |
    |                                                                                              |
    | Specifying *any* of ``act_as``, ``read_as``, ``admin``, ``ledger_id``, ``application_name``, |
    | ``oauth_token``, or ``oauth_token_file`` causes *all* of these environment variables to be   |
    | ignored                                                                                      |
    +----------------------+-----------------------------------------------------------------------+
    | ``act_as``           | ``DAML_LEDGER_ACT_AS`` (as a comma-separated list of parties)         |
    +----------------------+-----------------------------------------------------------------------+
    | ``read_as``          | ``DAML_LEDGER_READ_AS`` (as a comma-separated list of parties)        |
    +----------------------+-----------------------------------------------------------------------+
    | ``ledger_id``        | ``DAML_LEDGER_ID``                                                    |
    +----------------------+-----------------------------------------------------------------------+
    | ``application_name`` | ``DAML_LEDGER_APPLICATION_NAME``                                      |
    +----------------------+-----------------------------------------------------------------------+
    | ``oauth_token``      | ``DAML_LEDGER_OAUTH_TOKEN``                                           |
    +----------------------+-----------------------------------------------------------------------+
    | ``oauth_token_file`` | ``DAML_LEDGER_OAUTH_TOKEN_FILE``                                      |
    +----------------------+-----------------------------------------------------------------------+

    .. note::

        When connecting to the gRPC Ledger API, note that
        `gRPC environment variables
        <https://grpc.github.io/grpc/cpp/md_doc_environment_variables.html>`_ are also respected.
        You can configure the gRPC client in much more detail with these flags.

    :param url:
        The URL to connect to. Can be used as an alternative to supplying ``host``, ``port``,
        and ``scheme`` as individual values.

        If none of of ``url``, ``host``, ``port`` or ``scheme`` are specified, the value from the
        environment variable ``DAML_LEDGER_URL`` is used instead.

    :param host:
        The host to connect to. Can be used as an alternative to supplying ``url`` as a combined
        value.

        If none of of ``url``, ``host``, ``port`` or ``scheme`` are specified, the value from the
        environment variable ``DAML_LEDGER_HOST`` is used instead.

    :param port:
        The port to connect to. Can be used as an alternative to supplying ``url`` as a combined
        value.

        If none of of ``url``, ``host``, ``port`` or ``scheme`` are specified, the value from the
        environment variable ``DAML_LEDGER_PORT`` is used instead.

    :param scheme:
        The scheme to connect to. Can be used as an alternative to supplying ``url`` as a
        combined value.

        If none of of ``url``, ``host``, ``port`` or ``scheme`` are specified, the value from the
        environment variable ``DAML_LEDGER_SCHEME`` is used instead.

    :param connect_timeout:
        Length of time to wait before giving up connecting to the remote and declaring an error.
        The default value is 30 seconds.

    :param use_http_proxy:
        ``True`` to use an HTTP(S) proxy server if configured; ``False`` to avoid using any
        configured server. If unspecified and the host is ``localhost``, the proxy server is
        avoided; otherwise the proxy server is used.

    :param ca:
        A certificate authority to use to validate the server's certificate. If not supplied,
        the operating system's default trust store is used. Cannot be specified with
        ``ca_file``.

    :param ca_file:
        A file containing the certificate authority to use to validate the server's certificate.
        If not supplied, the operating system's default trust store is used. Cannot be specified
        with ``ca``.

    :param cert:
        A client-side certificate to be used when connecting to a server that requires mutual
        TLS. Cannot be specified with ``cert_file``.

    :param cert_file:
        A file containing the client-side certificate to be used when connecting to a server
        that requires mutual TLS. Cannot be specified with ``cert``.

    :param cert_key:
        A client-side private key to be used when connecting to a server that requires mutual
        TLS. Cannot be specified with ``cert_key_file``.

    :param cert_key_file:
        A client-side private key to be used when connecting to a server that requires mutual
        TLS. Cannot be specified with ``cert_key``.

    :param read_as:
        A party or set of parties on whose behalf (in addition to all parties listed in
        ``act_as``) contracts can be retrieved.

        Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.

        If none of of ``read_as``, ``act_as``, ``admin``, ``ledger_id``, ``application_name``,
        ``oauth_token``, or ``oauth_token_file`` are specified, the value from the environment
        variable ``DAML_LEDGER_ACT_AS`` is used instead.

    :param act_as:
        A party or set of parties on whose behalf commands should be executed. Parties here are
        also implicitly granted ``read_as`` access as well.

        Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.

        If none of of ``read_as``, ``act_as``, ``admin``, ``ledger_id``, ``application_name``,
        ``oauth_token``, or ``oauth_token_file`` are specified, the value from the environment
        variable ``DAML_LEDGER_ACT_AS`` is used instead.

    :param admin:
        HTTP JSON API only: allow admin endpoints to be used. This flag is ignored when
        connecting to gRPC Ledger API implementations.

        Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.

    :param ledger_id:
        The ledger ID to connect to. For the HTTP JSON API, this value is required. For the gRPC
        Ledger API, if this value is _not_ supplied, its value will be retrieved from the
        server.

        Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.

        If none of of ``read_as``, ``act_as``, ``admin``, ``ledger_id``, ``application_name``,
        ``oauth_token``, or ``oauth_token_file`` are specified, the value from the environment
        variable ``DAML_LEDGER_ID`` is used instead.

    :param application_name:
        A string that identifies this application. This is used for tracing purposes on the
        server-side.

        Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.

        If none of of ``read_as``, ``act_as``, ``admin``, ``ledger_id``, ``application_name``,
        ``oauth_token``, or ``oauth_token_file`` are specified, the value from the environment
        variable ``DAML_LEDGER_APPLICATION_NAME`` is used instead.

    :param oauth_token:
        The OAuth bearer token to be used on all requests.

        Cannot be specified if ``read_as``, ``act_as``, ``admin``, ``ledger_id``,
        ``application_name``, or ``oauth_token_file`` is specified.

        If none of of ``read_as``, ``act_as``, ``admin``, ``ledger_id``, ``application_name``,
        ``oauth_token``, or ``oauth_token_file`` are specified, the value from the environment
        variable ``DAML_LEDGER_OAUTH_TOKEN`` is used instead.

    :param oauth_token_file:
        A file that contains the OAuth bearer token to be used on all requests.

        Cannot be specified if ``read_as``, ``act_as``, ``admin``, ``ledger_id``,
        ``application_name``, or ``oauth_token`` is specified.

        If none of of ``read_as``, ``act_as``, ``admin``, ``ledger_id``, ``application_name``,
        ``oauth_token``, or ``oauth_token_file`` are specified, the value from the environment
        variable ``DAML_LEDGER_OAUTH_TOKEN_FILE`` is used instead.

    :param logger:
        The logger to use for connections created from this configuration. If not supplied, a
        logger will be created.

    :param logger_name:
        The name of the logger. Only used if ``logger`` is not provided.

    :param log_level:
        The logging level for the logger. The default is ``warn``. Only used if ``logger`` is
        not provided.
    """
    # TODO: Ideally we could _define_ the function as taking all of these parameters, but implement
    #  it with kwargs passing; unfortunately Sphinx doesn't seem to allow this.
    from .config import Config

    _ = Config.create(
        url=url,
        host=host,
        port=port,
        scheme=scheme,
        read_as=read_as,
        act_as=act_as,
        admin=admin,
        ledger_id=ledger_id,
        application_name=application_name,
        oauth_token=oauth_token,
        oauth_token_file=oauth_token_file,
        ca=ca,
        ca_file=ca_file,
        cert=cert,
        cert_file=cert_file,
        cert_key=cert_key,
        cert_key_file=cert_key_file,
        connect_timeout=connect_timeout,
        use_http_proxy=use_http_proxy,
        logger=logger,
        logger_name=logger_name,
        log_level=log_level,
    )
    # TODO: Implement when the v8 connection code fully lands
    raise NotImplementedError



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import re
from typing import Any, Optional

__all__ = [
    "LEDGER_STRING_REGEX",
    "NAME_STRING_REGEX",
    "PACKAGE_ID_STRING_REGEX",
    "PARTY_ID_STRING_REGEX",
    "to_bool",
    "to_str",
]

# Standard string regexes as defined here:
# https://github.com/digital-asset/daml/blob/a6da995ecb71004c34c88a4f4211543868cfde15/ledger-api/grpc-definitions/com/daml/ledger/api/v1/value.proto#L18-L21
NAME_STRING_REGEX = re.compile(r"[A-Za-z$_][A-Za-z0-9$_]*")
PACKAGE_ID_STRING_REGEX = re.compile(r"[A-Za-z0-9\-_ ]+")
PARTY_ID_STRING_REGEX = re.compile(r"[A-Za-z0-9:\-_ ]")
LEDGER_STRING_REGEX = re.compile(r"[A-Za-z0-9#:\-_/ ]")


def to_bool(obj: "Optional[Any]") -> "Optional[bool]":
    """
    Convert any of the common wire representations of a ``bool`` to a ``bool``.
    """
    if obj is None:
        # noinspection PyTypeChecker
        return None

    if isinstance(obj, bool):
        return obj
    elif isinstance(obj, str):
        lobj = obj.lower().strip()
        if lobj == "true":
            return True
        elif lobj == "false":
            return False
        else:
            raise ValueError(f"Could not parse a string as a boolean: {obj!r}")
    raise ValueError(f"Could not parse as a boolean: {obj!r}")


def to_str(obj: "Any") -> str:
    """
    Convert any object to a string. This simply calls ``str`` on the object to produce a string
    representation.
    """
    return obj if isinstance(obj, str) else str(obj)
